/*
**    TP 5ETI CPE Lyon
**    Copyright (C) 2011 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef _MATRIX3_HPP_
#define _MATRIX3_HPP_

#include <iostream>
#include <vector>

namespace cpe
{
    class v3;

    class matrix3
    {

    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor (identity) */
        matrix3();
        /** \brief direct constructor with every value */
        matrix3(const double& x00,const double& x01,const double& x02,
                const double& x10,const double& x11,const double& x12,
                const double& x20,const double& x21,const double& x22);

        // ********************************************* //
        // ********************************************* //
        //  Special initialization
        // ********************************************* //
        // ********************************************* //

        /** \brief build identity matrix */
        static matrix3 identity();
        /** \brief build zero matrix */
        static matrix3 zeros();
        /** \brief build rotation matrix */
        static matrix3 rotation(const v3& axis,const double& angle);
        /** \brief build scaling matrix */
        static matrix3 scale(const double& s);
        /** \brief build scaling matrix */
        static matrix3 scale(const double& s_x,const double& s_y,const double& s_z);


        // ********************************************* //
        // ********************************************* //
        //  Element access
        // ********************************************* //
        // ********************************************* //

        /** \brief Access to the k_th entry (k in [0,3]) */
        const double& operator()(const size_t& k1,const size_t& k2) const;
        /** \brief Access to the k_th entry (k in [0,3]) */
        double& operator()(const size_t& k1,const size_t& k2);

        /** \brief fast pointer access
            \warning unprotected access */
        const double* pointer() const;
        /** \brief fast pointer access
            \warning unprotected access */
        double* pointer_unprotected();



        // ********************************************* //
        // ********************************************* //
        //  Math operator
        // ********************************************* //
        // ********************************************* //

        /** \brief + operator */
        friend matrix3 operator+(const matrix3& m1,const matrix3& m2);
        /** \brief + operator */
        friend matrix3 operator+(const double& s,const matrix3& m);
        /** \brief + operator */
        friend matrix3 operator+(const matrix3& m,const double& s);

        /** \brief - operator */
        friend matrix3 operator-(const matrix3& m1,const matrix3& m2);
        /** \brief - operator */
        friend matrix3 operator-(const double& s,const matrix3& m);
        /** \brief - operator */
        friend matrix3 operator-(const matrix3& m,const double& s);

        /** \brief multiply by a scalar operator */
        friend matrix3 operator*(const double& s,const matrix3& m);
        /** \brief multiply by a scalar operator */
        friend matrix3 operator*(const matrix3& m,const double& s);
        /** \brief multiply matrix with a v3 */
        friend v3 operator*(const matrix3& m,const v3& v);
        /** \brief matrix mulitplication */
        friend matrix3 operator*(const matrix3& m1,const matrix3& m2);
        /** \brief divide by a scalar operator */
        friend matrix3 operator/(const matrix3& m,const double& s);


        /** \brief internal + */
        matrix3& operator+=(const matrix3& m);
        /** \brief internal + */
        matrix3& operator+=(const double& s);
        /** \brief internal - */
        matrix3& operator-=(const matrix3& m);
        /** \brief internal - */
        matrix3& operator-=(const double& s);
        /** \brief internal * */
        matrix3& operator*=(const double& s);
        /** \brief internal / */
        matrix3& operator/=(const double& s);

        /** \brief unary negation */
        matrix3 operator-() const;

        /** \brief does componentwise mutliplication
            \return MC_v2d(x*p.x(),y*p.y(),z*p.z())
        */
        matrix3 product_compontentwise(const matrix3& m) const;
        /** \brief does componentwise mutliplication

            perform p=(p.x()*p2.x(),p.y()*p2.x())
        */
        matrix3& product_compontentwise_internal(const matrix3& m);




        // ********************************************* //
        // ********************************************* //
        //  Output
        // ********************************************* //
        // ********************************************* //

        /** \brief output the vector in ostream */
        friend std::ostream& operator<<(std::ostream& stream,const matrix3& m);




        // ********************************************* //
        // ********************************************* //
        //  Projection matrix
        // ********************************************* //
        // ********************************************* //

        /** \brief convert 3x3 matrix into a 4x4 vector of double (glMultmatrixf) */
        std::vector<double> to_4d() const;

    private:

        // ********************************************* //
        // ********************************************* //
        //  Internal parameters
        // ********************************************* //
        // ********************************************* //

        /** \brief internal storage of the matrix */
        double m[9];


        // ********************************************* //
        // ********************************************* //
        //  Helper
        // ********************************************* //
        // ********************************************* //

        /** \brief assert that a size_t belongs to [[0,3]] */
        void assert_size(const size_t& k1,const size_t& k2) const throw(std::exception);

    };
}

#endif
